# FFTSearch: Genarating and Searching Families of FFT Algorithms
# Copyright (C) 2011 Steve Haynal, Heidi Haynal
#
# This file is part of FFTSearch.
#
# FFTSearch is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFTSearch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from ModInt import *

import math

## Container class for all BitVector types in a node
class BV:
	def __init__(self,c,id,sfgsz=0):
		self.c = c
		self.sid = "%d_%d_%d" % id
		if sfgsz != 0:
			self.sid = "%s_%d" % (self.sid,sfgsz)

	def __setattr__(self, name, value):
		if isinstance(value,BitVector):
			sname = "%s_%s" % (name,self.sid)
			if isinstance(value,Bit):
				v = NamedBit(sname,[value])
			else:
				v = NamedVec(sname,[value])
			self.__dict__[name] = v
			if value.IsExternal():
				self.c.external.append(v)
			else:
				self.c.internal.append(v)
		else:
			self.__dict__[name] = value

		
class BitVector:
	
	def __init__(self,children):
		self.children = children
		self.parents = []

	def __str__(self):
		if self.children != []:
			s = ','.join(str(n) for n in self.children)
			return "%s[%s])" % (self.Op(),s)
		else:
			return "%s)" % self.Op()
			
	def SMTLib1_2(self):
		if self.children != []:
			s = ' '.join(n.SMTLib1_2() for n in self.children)
			return "(%s %s)" % (self.Op_SMTLib1_2(),s)
		else:
			return "(%s)" % self.Op_SMTLib1_2()
			
	def QFBV(self): 
		if self.children != []:
			s = ' '.join(n.QFBV() for n in self.children)
			return "(%s %s)" % (self.Op_QFBV(),s)
		else:
			return "(%s)" % self.Op_QFBV()	
	
	def LIA(self):
		op = self.Op_LIA()
		if self.children != []:
			s = ' '.join(n.LIA() for n in self.children)
			if op == "":
				return s
			else:
				return "(%s %s)" % (op,s)
		else:
			if op == "":
				return ""
			else:
				return "(%s)" % op			
						
		
	def IsExternal(self): return False
	
	def MarkParents(self,p):
		if self.parents == []:
			for c in self.children:
				c.MarkParents(self)
		self.parents.append(p)

## Bit	
class Bit(BitVector):
	width = 1
	def __init__(self,children):
		BitVector.__init__(self,children)

	def __repr__(self): return "Bit()"

	def __and__(self,other): return And([self,other])
	def __or__(self,other): return Or([self,other])	
	def __xor__(self,other): return Xor([self,other])		
	def __invert__(self): return Invert(self)
	 
class Bit0(Bit):
	def __init__(self):
		Bit.__init__(self,[])

	def SMTLib1_2(self): return "bit0"
	def QFBV(self): return "(_ bv0 1)"
	def Op(self): return "Bit0("
	def __repr__(self): return "Bit0()"

class Bit1(Bit):
	def __init__(self):
		Bit.__init__(self,[])

	def SMTLib1_2(self): return "bit1"
	def QFBV(self): return "(_ bv1 1)"
	def Op(self): return "Bit1("
	def __repr__(self): return "Bit1()"
	
class ExtBit(Bit):
	def __init__(self):
		Bit.__init__(self,[])

	def Op(self): return "ExtBit("
	def __repr__(self): return "ExtBit()"
	
	def IsExternal(self): return True

class NamedBit(Bit):
	def __init__(self,name,children):
		Bit.__init__(self,children)
		self.name  = name

	def Op(self): return "NamedBit(%s,%d," % (self.name,len(self.parents))
	def __str__(self): return self.name
	def __repr__(self): return "NamedBit()"

	def AssignmentStr(self):
		return "%s = %s" % (self,self.children[0])


	def SMTLib1_2(self):
		c = self.children[0]
		if isinstance(c,ExtBit):
			return self.name
		else:
			return "$%s" % self.name		

	def AssignmentStr_SMTLib1_2(self):
		c = self.children[0]
		if isinstance(c,ExtBit):
			s = ":extrapreds ((%s))" % self.name
		else:
			s = "(flet ($%s %s)" % (self.name,self.children[0].SMTLib1_2())
		return s
		
	def LIA(self):
		c = self.children[0]
		if isinstance(c,ExtBit):
			return self.name
		else:
			return "?%s" % self.name
			
	def QFBV(self): return self.LIA()			

	def AssignmentStr_LIA(self):
		c = self.children[0]
		if isinstance(c,ExtBit):
			s = "(declare-fun %s () Bool)" % self.name
		else:
			s = "(let ((?%s %s))" % (self.name,self.children[0].LIA())
		return s		
		
	def AssignmentStr_QFBV(self):
		c = self.children[0]
		if isinstance(c,ExtBit):
			s = "(declare-fun %s () Bool)" % self.name
		else:
			s = "(let ((?%s %s))" % (self.name,self.children[0].QFBV())
		return s			
	

class Equality(Bit):
	def Op_SMTLib1_2(self): return "="
	def Op_LIA(self): return "="
	def Op_QFBV(self): return self.Op_SMTLib1_2()
	def Op(self): return "Equality("
	def __repr__(self): return "Equality()"

class ULE(Bit):
	def Op_SMTLib1_2(self): return "bvule"
	def Op_LIA(self): return "<="
	def Op_QFBV(self): return self.Op_SMTLib1_2()	
	def Op(self): return "ULE("
	def __repr__(self): return "ULE()"
	
class UL(Bit):
	def Op_SMTLib1_2(self): return "bvult"
	def Op_LIA(self): return "<"	
	def Op_QFBV(self): return self.Op_SMTLib1_2()	
	def Op(self): return "UL("
	def __repr__(self): return "UL()"	


class And(Bit):
	def Op_SMTLib1_2(self): return "and"	
	def Op_LIA(self): return "and"
	def Op_QFBV(self): return self.Op_SMTLib1_2()	
	def Op(self): return "And("	
	def __repr__(self): return "And()"	
		
class Or(Bit):
	def Op_SMTLib1_2(self): return "or"	
	def Op_QFBV(self): return self.Op_SMTLib1_2()	
	def Op(self): return "Or("
	def Op_LIA(self): return "or"
	def __repr__(self): return "Or()"			

class Xor(Bit):
	def Op_SMTLib1_2(self): return "xor"	
	def Op_QFBV(self): return self.Op_SMTLib1_2()	
	def Op(self): return "Xor("	
	def __repr__(self): return "Xor()"	

class Invert(Bit):
	def __init__(self,r):
		Bit.__init__(self,[r])

	def Op_SMTLib1_2(self): return "not"
	def Op_LIA(self): return "not"
	def Op_QFBV(self): return self.Op_SMTLib1_2()	
	def Op(self): return "~("
	def __repr__(self): return "Invert()"	
	
	
		
class Vector(BitVector):
	def __init__(self,children,width=None):
		BitVector.__init__(self,children)
		if width == None:
			self.width = children[0].width
		else:
			self.width = width
		
	def __repr__(self): return "Vector()"
		
	def __add__(self,other):
		## Reduce add with constant zero
		if isinstance(other,ConstVec):
			if int(other.v) == 0: return self
		elif isinstance(self,ConstVec):
			if int(self.v) == 0: return other

		assert self.width == other.width, 'ERROR: Vectors with different widths.'
		return Add([self,other])
		
	def __sub__(self,other):
		## Reduce sub with RHS of constant zero
		if isinstance(other,ConstVec):
			if int(other.v) == 0: return self
		
		assert self.width == other.width, 'ERROR: Vectors with different widths.'
		return Sub([self,other])
		
	def __eq__(self,other):
		assert self.width == other.width, 'ERROR: Vectors with different widths.'
		return Equality([self,other])		
	
	def __le__(self,other): 
		assert self.width == other.width, 'ERROR: Vectors with different widths.'
		return ULE([self,other])	
		
	def __lt__(self,other): 
		assert self.width == other.width, 'ERROR: Vectors with different widths.'
		return UL([self,other])			
				


class Ite(Vector):
	def __init__(self,s,t,e):
		BitVector.__init__(self,[s,t,e])
		self.s = s
		assert t.width == e.width, 'ERROR: Vectors with different widths.'
		self.width = t.width
		
		
	def Op_SMTLib1_2(self): return "ite"
	def Op_LIA(self): return "ite"
	def Op_QFBV(self): return self.Op_SMTLib1_2()		
	def Op(self): return "ITE(%s," % self.s
	def __repr__(self): return "Ite()"	
	
	
class ConstVec(Vector):
	def __init__(self,v,width=None):
		if width == None:
			width = int(math.log(v.mod,2))	
		Vector.__init__(self,[],width)
		self.v = int(v)

	def SMTLib1_2(self): return "bv%d[%d]" % (self.v,self.width)
	def QFBV(self): return "(_ bv%d %d)" % (self.v,self.width)
	def LIA(self):
		if self.v < 0:
			return "(- %d)" % abs(self.v)
		else: 
			return "%d" % self.v
	def Op(self): return "ConstVec(%d" % self.v
	def __repr__(self): return "ConstVec(%d)" % self.v

def C(v,w=None):
	return ConstVec(v,w)

class ExtVec(Vector):
	def __init__(self,width):
		Vector.__init__(self,[],width)

	def Op(self): return "ExtVec("
	def __repr__(self): return "ExtVec()"
	
	def IsExternal(self): return True


class NamedVec(Vector):
	def __init__(self,name,children):
		Vector.__init__(self,children)
		self.name = name

	def Op(self): return "NamedVec(%s,%d," % (self.name,len(self.parents))
	def __str__(self): return self.name	
	def __repr__(self): return "NamedVec()"
	
	def AssignmentStr(self):
		return "%s = %s" % (self,self.children[0])
		
		
	def SMTLib1_2(self):
		c = self.children[0]
		if isinstance(c,ExtVec):
			return self.name
		else:
			return "?%s" % self.name

		
		
	def AssignmentStr_SMTLib1_2(self):
		c = self.children[0]
		if isinstance(c,ExtVec):
			s = ":extrafuns ((%s BitVec[%d]))" % (self.name,self.width)
		else:
			s = "(let (?%s %s)" % (self.name,self.children[0].SMTLib1_2())
		return s

	def QFBV(self): return self.SMTLib1_2()

	def AssignmentStr_QFBV(self):
		c = self.children[0]
		if isinstance(c,ExtVec):
			s = "(declare-fun %s () (_ BitVec %d))" % (self.name,self.width)
		else:
			s = "(let ((?%s %s))" % (self.name,self.children[0].QFBV())
		return s


	def LIA(self):
		c = self.children[0]
		if isinstance(c,ExtVec):
			return self.name
		else:
			return "?%s" % self.name		
		
	def AssignmentStr_LIA(self):
		c = self.children[0]
		if isinstance(c,ExtVec):
			s = "(declare-fun %s () Int)(assert (< %s %d))(assert (>= %s 0))" % (self.name,self.name,2**self.width,self.name)
		else:
			s = "(let ((?%s %s))" % (self.name,self.children[0].LIA())
		return s					



class Add(Vector):
	def Op_SMTLib1_2(self): return "bvadd"
	def Op_LIA(self): return "+"
	def Op_QFBV(self): return self.Op_SMTLib1_2()
	def Op(self): return "Add("
	def __repr__(self): return "Add()"	
	
class ZeroExtend(Vector):
	def __init__(self,child,width):
		Vector.__init__(self,[child],width)	
		
	def Op_SMTLib1_2(self): return "zero_extend[%d]" % (self.width - self.children[0].width)
	def Op_LIA(self): return ""	
	def Op_QFBV(self): return "(_ zero_extend %d)" % (self.width - self.children[0].width)
	def Op(self): return "ZeroExtend("
	def __repr__(self): return "ZeroExtend()"			
		
	def SMTLib1_2(self):
		if self.children[0].width == self.width:
			return self.children[0].SMTLib1_2()
		else:
			s = ' '.join(n.SMTLib1_2() for n in self.children)
			return "(%s %s)" % (self.Op_SMTLib1_2(),s)

	def QFBV(self):
		if self.children[0].width == self.width:
			return self.children[0].QFBV()
		else:
			s = ' '.join(n.QFBV() for n in self.children)
			return "(%s %s)" % (self.Op_QFBV(),s)
		
class Sub(Vector):
	def Op_SMTLib1_2(self): return "bvsub"	
	def Op_QFBV(self): return self.Op_SMTLib1_2()
	def Op_LIA(self): return "-"	
	def Op(self): return "Sub("	
	def __repr__(self): return "Sub()"	
	
class Concat(Vector):
	def __init__(self,l,r):
		Vector.__init__(self,[l,r],l.width+r.width)	
	
	def Op_SMTLib1_2(self): return "concat"
	def Op_QFBV(self): return self.Op_SMTLib1_2()			
	def Op(self): return "Concat("	
	def __repr__(self): return "Concat()"		
			
class Extract(Vector):
	def __init__(self,(hi,lo),r):
		self.hi = hi
		self.lo = lo
		Vector.__init__(self,[r],(hi-lo)+1)

	def Op_SMTLib1_2(self): return "extract[%d:%d]" % (self.hi,self.lo)
	def Op_QFBV(self): return "(_ extract %d %d)" % (self.hi,self.lo)
	def Op(self): return "Extract(%d:%d," % (self.hi,self.lo)
	def __repr__(self): return "Extract((%d,%d))" % (self.hi,self.lo)			
		
			

	

