# FFTSearch: Genarating and Searching Families of FFT Algorithms
# Copyright (C) 2011 Steve Haynal, Heidi Haynal
#
# This file is part of FFTSearch.
#
# FFTSearch is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFTSearch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from BitVector import *

import sys
sys.setrecursionlimit(10000)

## Base class for constraints
class Constraints:
	
	def __init__(self,width,fg):
		
		self.width = width
		self.N = 2**width
		self.fg = fg
		self.good = []
		self.bad = []
		self.external = []
		self.internal = []
		
		## Init BV 
		for nd in self.fg:
			nd.bv = BV(self,nd.id)	
			
		for sfg in self.fg.AllSFGs():
			sfg.bv = BV(self,sfg.inps[0].id,sfg.m)

	def BuildMaxCost(self,includetop=True):
		'''Applies a max FLOP count constraint'''
		
		if self.maxcost == 0: return
		
		maxbits = int(math.ceil(math.log((2*self.maxcost)+1,2)))	
		
		def MaxToBits(n):
			sz = int(math.ceil(math.log(n+1,2)))
			if sz > maxbits: return maxbits
			else: return sz

		def BuildICost(sfg,maxcost):
			if sfg.N > 2:
				BuildICost(sfg.tl,maxcost)
				BuildICost(sfg.tr,maxcost)
				
			sfg.bv.maxicost = sfg.tl.bv.maxicost + sfg.tr.bv.maxicost
			isz = MaxToBits(sfg.bv.maxicost)
			sfg.bv.icost = ZeroExtend(sfg.tl.bv.icost,isz) + ZeroExtend(sfg.tr.bv.icost,isz)
			if sfg.bv.maxicost > maxcost:
				sfg.bv.imaxcost = sfg.bv.icost <= ConstVec(maxcost,isz)
				self.good.append(sfg.bv.imaxcost)
			
		for sfg in self.fg.SFGsOut2In():
			maxcost = self.maxcost				

			sfg.bv.maxbcost = sfg.bl.bv.maxtcost + sfg.br.bv.maxtcost
			bsz = MaxToBits(sfg.bv.maxbcost)
			sfg.bv.bcost = ZeroExtend(sfg.bl.bv.tcost,bsz) + ZeroExtend(sfg.br.bv.tcost,bsz)
			if sfg.bv.maxbcost > maxcost:
				sfg.bv.bmaxcost = sfg.bv.bcost <= ConstVec(maxcost,bsz)
				self.good.append(sfg.bv.bmaxcost)
			
			if self.fg != sfg or includetop:
			
				BuildICost(sfg,maxcost)			
			
				sfg.bv.maxtcost = sfg.bv.maxbcost + sfg.bv.maxicost
				tsz = MaxToBits(sfg.bv.maxtcost)
				sfg.bv.tcost = ZeroExtend(sfg.bv.icost,tsz) + ZeroExtend(sfg.bv.bcost,tsz)						

				if sfg.bv.maxtcost > maxcost:
					sfg.bv.tmaxcost = sfg.bv.tcost <= ConstVec(maxcost,tsz)
					self.good.append(sfg.bv.tmaxcost)


	def MarkParents(self):
		'''Mark expression graph to share expressions'''
		for nd in self.good:
			nd.MarkParents(None)
		for nd in self.bad:
			nd.MarkParents(None)

				
	def WriteSMTLib1_2(self,fn):
		'''Generates constraints in SMT lib 1.2 format'''
		
		self.f = open(fn,"w")
		self.f.write("(benchmark fftgraph%d\n" % self.fg.m)
		self.f.write(":source { Steve Haynal }\n")
		self.f.write(":category { crafted }\n")
		self.f.write(":status unknown\n")
		self.f.write(":logic QF_BV\n")
			
		for nd in self.external:
			self.f.write(nd.AssignmentStr_SMTLib1_2())
			self.f.write('\n')
			
		self.f.write(":formula\n")
		
		parens = 0
		for nd in self.internal:
			if nd.parents != []:
				self.f.write(nd.AssignmentStr_SMTLib1_2())
				self.f.write("\n")
				parens += 1
				
		## Good
		self.f.write("(flet ($good (and true")
		i = 0
		for nd in self.good:
			self.f.write(" ")
			self.f.write(nd.SMTLib1_2())
			if i == 5:
				i = 0
				self.f.write("\n")
			else:
				i += 1
		self.f.write("))\n")
		parens +=1
		
		## Bad
		self.f.write("(flet ($bad (or false")
		i = 0
		for nd in self.bad:
			self.f.write(" ")
			self.f.write(nd.SMTLib1_2())
			if i == 5:
				i = 0
				self.f.write("\n")
			else:
				i += 1	
		self.f.write("))\n")
		parens += 1
				
		self.f.write("(and $good (not $bad))\n")
		
		while parens > 30:
			self.f.write("))))))))))))))))))))))))))))))\n")				
			parens = parens - 30
		if parens > 0:
			s = ")" * parens
			self.f.write(s)
			self.f.write("\n")
		self.f.write(")\n")
		self.f.close()					


	def WriteLIA(self,fn):
		'''Generates constraints in SMT QF_LIA 2.0 format'''
		
		self.f = open(fn,"w")
		self.f.write("(set-logic QF_LIA)\n")
		self.f.write("(set-info :source |\nSearch for n FLOP solution to size-n FFT\nSteve Haynal <steve@softerhardware.com>\n|)\n")
		self.f.write("(set-info :smt-lib-version 2.0)\n")
		self.f.write('(set-info :category "crafted")\n')
		self.f.write("(set-info :status unknown)\n")
			
		for nd in self.external:
			self.f.write(nd.AssignmentStr_LIA())
			self.f.write('\n')
			
		self.f.write("(assert\n")
		
		parens = 0
		for nd in self.internal:
			if nd.parents != []:
				self.f.write(nd.AssignmentStr_LIA())
				self.f.write("\n")
				parens += 1
		
		## Good
		self.f.write("(and true")
		i = 0
		for nd in self.good:
			self.f.write(" ")
			self.f.write(nd.QFBV())
			if i == 5:
				i = 0
				self.f.write("\n")
			else:
				i += 1
		self.f.write(")\n")
		parens +=1
				
		while parens > 30:
			self.f.write("))))))))))))))))))))))))))))))\n")				
			parens = parens - 30
		if parens > 0:
			s = ")" * parens
			self.f.write(s)
			self.f.write("\n")
			
		self.f.write("(check-sat)\n")
		self.f.write("(exit)\n")
		self.f.close()		
							


	def WriteQFBV(self,fn):
		'''Generates constraints in SMT QF_BV 2.0 format'''
		
		self.f = open(fn,"w")
		self.f.write("(set-logic QF_BV)\n")
		self.f.write("(set-info :source |\nSearch for n FLOP solution to size-n FFT\nSteve Haynal <steve@softerhardware.com>\n|)\n")		
		self.f.write("(set-info :smt-lib-version 2.0)\n")
		self.f.write('(set-info :category "crafted")\n')
		self.f.write("(set-info :status unknown)\n")
			
		for nd in self.external:
			self.f.write(nd.AssignmentStr_QFBV())
			self.f.write('\n')
			
		self.f.write("(assert\n")
		
		parens = 0
		for nd in self.internal:
			if nd.parents != []:
				self.f.write(nd.AssignmentStr_QFBV())
				self.f.write("\n")
				parens += 1

		## Good
		self.f.write("(let ((?good (and true")
		i = 0
		for nd in self.good:
			self.f.write(" ")
			self.f.write(nd.LIA())
			if i == 5:
				i = 0
				self.f.write("\n")
			else:
				i += 1
		self.f.write(")))\n")				
		parens +=1
		
		## Bad
		self.f.write("(let ((?bad (or false")
		i = 0
		for nd in self.bad:
			self.f.write(" ")
			self.f.write(nd.LIA())
			if i == 5:
				i = 0
				self.f.write("\n")
			else:
				i += 1	
		self.f.write(")))\n")
		parens += 1
				
		self.f.write("(and ?good (not ?bad)))\n")
		
		while parens > 30:
			self.f.write("))))))))))))))))))))))))))))))\n")				
			parens = parens - 30
		if parens > 0:
			s = ")" * parens
			self.f.write(s)
			self.f.write("\n")
			
		self.f.write("(check-sat)\n")
		self.f.write("(exit)\n")
		self.f.close()		


## Base class for contraints built on size-2 butterflies
class BFConstraints(Constraints):
			
	def __init__(self,fg,maxcost=0):
		sm = int(math.log(fg.inps[0].Wbase.mod,2)) - 2
		Constraints.__init__(self,sm,fg)		
		
		if maxcost > 0:
			adds = 2 * fg.m * fg.N
			self.maxcost = (maxcost - adds) / 2
			print "Adjusted maxcost is",self.maxcost
		else:
			self.maxcost = maxcost
			
			
	def DoMaxCost(self):
		'''Build constraints as described in paper for a maximum FLOP count'''
		self.BuildBFWbaseTFs()
		self.BuildNodeSharedTFCostsWithPredicates()
		self.PinSize2()
		self.BuildPushUp()
		self.BuildMaxCost()
		self.MarkParents()
		

	def DoLIA(self):
		'''Build constraints as described in paper for a maximum FLOP count'''
		self.BuildLIABFWbaseTFs()
		self.BuildNodeSharedTFCostsLIA()		
		self.PinSize2()
		self.BuildPushUp()
		self.BuildMaxCost()
		self.MarkParents()	


	def DoRestrictTFs(self,constraints):
		'''Build constraints as described in paper for restricted set of twiddle factors'''
		self.BuildBFWbaseTFs()
		self.BuildNodeSharedTFCostsWithPredicates()	
		
		## Do restriction		
		for nd in self.fg:
			rtfs = nd.bv.ltfp == ConstVec(0,self.width)
			for c in constraints:
				rtfs = (nd.bv.ltfp == c) | rtfs
			nd.bv.rtfs = rtfs
			self.good.append(nd.bv.rtfs)

		self.BuildMaxCost()
		self.MarkParents()	




	def BuildBFWbaseTFs(self):
		'''Declare Wbase and TwiddleFactors'''
		
		## Handle terminal case at top
		for nd in self.fg.inps:
			Wbase = ConstVec(nd.Wbase % self.N)
			nd.bv.Wbase = Wbase
			nd.bv.rWbase = Wbase
			
		## Build Wbase
		for bf in self.fg.Butterflies():
			bf.bl.bv.Wbase = ExtVec(self.width)
			bf.bl.bv.rWbase = bf.bl.bv.Wbase + ConstVec(bf.bl.Wstride % self.N)
			bf.br.bv.Wbase = bf.bl.bv.Wbase
			bf.br.bv.rWbase = bf.bl.bv.rWbase
			
		## Build TF
		for nd in self.fg:
			if nd in self.fg.outs:
				if nd.IsTerminalDescendent():
					## Terminal descendent must have zero weight on a0
					nd.bv.ltfp = ConstVec(0,self.width) - nd.bv.Wbase
				else:
					## Terminal node weight plus tfp must match weight of child
					if nd.lc.lp == nd:
						nd.bv.ltfp = ConstVec(nd.lc.Wbase % self.N) - nd.bv.Wbase
					else:
						nd.bv.ltfp = ConstVec(nd.lc.rWbase() % self.N) - nd.bv.Wbase	
			else:
				if nd.lc.lp == nd:
					nd.bv.ltfp = nd.lc.bv.Wbase - nd.bv.Wbase
				else:
					nd.bv.ltfp = nd.lc.bv.rWbase - nd.bv.Wbase		

	
	def BuildLIABFWbaseTFs(self):
		'''Declare Wbase and TwiddleFactors'''
		
		liaN = ConstVec(self.fg.inps[0].Wbase % self.N)
		liaN.v = self.N
		liaZ = ConstVec(self.fg.inps[0].Wbase % self.N)
		liaZ.v = 0
		
		## Handle terminal case at top
		for nd in self.fg.inps:
			Wbase = ConstVec(nd.Wbase % self.N)
			nd.bv.Wbase = Wbase
			nd.bv.rWbase = Wbase
			
		## Build Wbase
		for bf in self.fg.Butterflies():
			bf.bl.bv.Wbase = ExtVec(self.width)
			bf.bl.bv.trWbase = bf.bl.bv.Wbase + ConstVec(bf.bl.Wstride % self.N)
			bf.bl.bv.rWbase = Ite( (bf.bl.bv.trWbase <= liaN),bf.bl.bv.trWbase,(bf.bl.bv.trWbase - liaN))
			bf.br.bv.Wbase = bf.bl.bv.Wbase
			bf.br.bv.rWbase = bf.bl.bv.rWbase
			
		## Build TF
		for nd in self.fg:
			if nd in self.fg.outs:
				if nd.IsTerminalDescendent():
					## Terminal descendent must have zero weight on a0
					nd.bv.tltfp = ConstVec(0,self.width) - nd.bv.Wbase
				else:
					## Terminal node weight plus tfp must match weight of child
					if nd.lc.lp == nd:
						nd.bv.tltfp = ConstVec(nd.lc.Wbase % self.N) - nd.bv.Wbase
					else:
						nd.bv.tltfp = ConstVec(nd.lc.rWbase() % self.N) - nd.bv.Wbase	
			else:
				if nd.lc.lp == nd:
					nd.bv.tltfp = nd.lc.bv.Wbase - nd.bv.Wbase
				else:
					nd.bv.tltfp = nd.lc.bv.rWbase - nd.bv.Wbase		
			nd.bv.ltfp = Ite( (nd.bv.tltfp < liaZ),(nd.bv.tltfp + liaN),nd.bv.tltfp)
	
					
	def BuildNodeSharedTFCostsWithPredicates(self):
		'''Declare cost of each node using predicates'''
		for nd in self.fg:
			nd.bv.lz = Extract((self.width-2,0),nd.bv.ltfp) == ConstVec(0,self.width-1)
			nd.bv.msbz = Extract((self.width-1,self.width-1),nd.bv.ltfp) == ConstVec(0,1)
			nd.bv.c0 = nd.bv.lz & nd.bv.msbz
			nd.bv.nc0 = ~nd.bv.c0
			nd.bv.c4 = nd.bv.lz & ~nd.bv.msbz
			nd.bv.c6 = ~nd.bv.lz
			
			nd.bv.icost = Concat(Ite(nd.bv.nc0,Bit1(),Bit0()),Ite(nd.bv.c6,Bit1(),Bit0()))
			nd.bv.tcost = nd.bv.icost
			nd.bv.maxicost = 3
			nd.bv.maxtcost = 3
	
	def BuildNodeSharedTFCosts(self):
		'''Declare cost of each node without predicates'''
		for nd in self.fg:
			lz = Extract((self.width-2,0),nd.bv.ltfp) == ConstVec(0,self.width-1)
			msbz = Extract((self.width-1,self.width-1),nd.bv.ltfp) == ConstVec(0,1)
			nd.bv.icost = Ite( lz, Ite(msbz,ConstVec(0,2),ConstVec(2,2)),ConstVec(3,2))
			
			nd.bv.tcost = nd.bv.icost
			nd.bv.maxicost = 3
			nd.bv.maxtcost = 3	
				
	def BuildNodeSharedTFCostsLIA(self):
		'''Declare cost of each node as ints'''
		for nd in self.fg:
			nd.bv.c0 = nd.bv.ltfp == ConstVec(0,self.width)
			nd.bv.c4 = nd.bv.ltfp == ConstVec(self.N/2,self.width)
			nd.bv.c6 = ~(nd.bv.c0 | nd.bv.c4)
			nd.bv.nc0 = ~nd.bv.c0
			
			nd.bv.icost = Ite( nd.bv.c0, ConstVec(0,2), Ite(nd.bv.c4,ConstVec(2,2),ConstVec(3,2)))
			
			nd.bv.tcost = nd.bv.icost
			nd.bv.maxicost = 3
			nd.bv.maxtcost = 3			

	def BuildNodeSharedTFCostsLIAFullRange(self):
		'''Declare cost of each node as ints'''
		liaN = ConstVec(0,self.width)
		liaN.v = self.N
		liaZ = ConstVec(0,self.width)
		liaZ.v = 0
		liaH0 = ConstVec(0,self.width)
		liaH0.v = self.N/2
		liaH1 = ConstVec(0,self.width)
		liaH1.v = 0 - (self.N/2)
		liaH2 = ConstVec(0,self.width)
		liaH2.v = self.N + (self.N/2)
				
		for nd in self.fg:
			c0 = Or([(nd.bv.ltfp == liaZ),(nd.bv.ltfp == liaN)])
			c4 = Or([(nd.bv.ltfp == liaH0),(nd.bv.ltfp == liaH1),(nd.bv.ltfp == liaH2)])
			nd.bv.icost = Ite( c0, ConstVec(0,2), Ite(c4,ConstVec(2,2),ConstVec(3,2)))
			
			nd.bv.tcost = nd.bv.icost
			nd.bv.maxicost = 3
			nd.bv.maxtcost = 3		



				
	def BuildPushDown(self):
		'''Apply top equal-pair symmetries'''
		for bf in self.fg.Butterflies():
			bf.bv.pushdown = bf.tl.bv.nc0 & (bf.tl.bv.ltfp == bf.tr.bv.ltfp)
			self.bad.append(bf.bv.pushdown)

	def BuildPushUp(self):
		'''Apply bottom equal-pair symmetries'''
		for bf in self.fg.Butterflies():
			bf.bv.pushup = bf.bl.bv.nc0 & (bf.bl.bv.ltfp == bf.br.bv.ltfp)
			self.bad.append(bf.bv.pushup)
			
	def PinSize2(self):
		'''Apply 3-Node symmetry constraints'''
		for bf in self.fg.Butterflies():
			bf.bv.pinzero = bf.bl.bv.c6 & ((bf.tl.bv.c6 & bf.tr.bv.c6) | (bf.tl.bv.c6 & bf.br.bv.c6) | (bf.tr.bv.c6 & bf.br.bv.c6))
			self.bad.append(bf.bv.pinzero)
			

			
