# FFTSearch: Genarating and Searching Families of FFT Algorithms
# Copyright (C) 2011 Steve Haynal, Heidi Haynal
#
# This file is part of FFTSearch.
#
# FFTSearch is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFTSearch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import pickle
from ModInt import *

## Simple operation parse tree for construction of FFT algorithms
class Op:
	def __init__(self):
		return
	
	def __add__(self,other):
		return Add(self,other)
				
class Add(Op):
	def __init__(self,l,r):
		self.l = l
		self.r = r
		self.id = None
		self.Wbase = None

class Omega(Op):
	def __init__(self,k,N,nd):
		self.W = ModInt(k,N)
		self.nd = nd
		
class Leaf(Op):
	def __init__(self,index):
		self.index = index

def omega((k,N),nd):
	return Omega(k,N,nd)
		
def negate(N,nd):
	return Omega(N >> 1,N,nd)
	
def img(N,nd):
	return Omega(N >> 2,N,nd)

def LoadFFTAlgorithm(fn):
	f = open(fn,'r')
	ffta = pickle.load(f)
	f.close()
	return ffta

## FFT Algorithm 
class FFTAlgorithm:
	def __init__(self,m):
		self.weights = {}
		self.N = (2**m)
		self.m = m
		
	def SaveFFTAlgorithm(self,fn):
		f = open(fn,'w')
		pickle.dump(self,f)
		f.close()		
		
	def ImportParseTree(self,nds):
		'''Helper function to build FFT algorithm from simple operation graph'''
		def Do(nd):
			if isinstance(nd,Add):
				if nd.id != None: return (nd.id,nd.Wbase)
			
				(lid,lweight) = Do(nd.l)
				(rid,rweight) = Do(nd.r)
		
				if lid[1] > rid[1]:
					tmp = rid; rid = lid; lid = tmp
					tmp = rweight; rweight = lweight; lweight = tmp
			
				lbase = lid[1]
				rbase = rid[1]
				nd.id = (rbase-lbase,lbase,int(rweight-lweight))
				nd.Wbase = lweight
				##print nd.id
				if nd.id in self.weights:
					print "WARNING: id",nd.id,"is already in weights"
				else:
					self.weights[nd.id] = nd.Wbase
				return (nd.id,nd.Wbase)
				
			elif isinstance(nd,Omega):
				(id,weight) = Do(nd.nd)
				return (id,weight.NMod(self.N) + nd.W.NMod(self.N))
			
			elif isinstance(nd,Leaf):
				id = (self.N,nd.index,0)
				Wbase = ModInt(0,self.N)
				self.weights[id] = Wbase
				return (id,Wbase)				
	
		return [Do(nd) for nd in nds]


	def ConjugateSplitRadix(self):
		'''Build conjugate split-radix algorithm
			subtraction does not always occur for bottom right child in size-2 butterfly.
		'''
		def Algorithm(xn):	
			N = len(xn)
			if N == 1:
				return xn
			elif N == 2:
				return [xn[0] + xn[1],xn[0] + negate(N,xn[1])]

			uk2 = Algorithm(xn[0::2])
			zk4 = Algorithm(xn[1::4])
			nxn = xn[3::4]
			nxn = nxn[-1:] + nxn[:-1]
			zpk4 = Algorithm(nxn)
		
			yk0 = []; yk1 = []; yk2 = []; yk3 = []
		
			## omega k N is defined as omega 2*pi*k here, not omega -2*pi*k
			## as is seen often in FFT algorithms
			for k in range(0,N>>2):
				t0 = omega((k,N),zk4[k])
				t1 = omega((N-k,N),zpk4[k])
				s0 = t0 + t1
				s1 = img(N,t0 + negate(N,t1))
			
				yk0.append(uk2[k] + s0)
				yk2.append(uk2[k] + negate(N,s0))
				yk1.append(uk2[k+(N>>2)] + s1)
				yk3.append(uk2[k+(N>>2)] + negate(N,s1))
			
			return yk0+yk1+yk2+yk3
		
		self.ImportParseTree(Algorithm([Leaf(i) for i in range(0,self.N)]))

	def SplitRadix(self):
		'''Build standard split-raidx FFT algorithm'''
		
		def Algorithm(xn):	
			N = len(xn)
			if N == 1:
				return xn
			elif N == 2:
				return [xn[0] + xn[1],xn[0] + negate(N,xn[1])]

			uk2 = Algorithm(xn[0::2])
			zk4 = Algorithm(xn[1::4])
			nxn = xn[3::4]
			zpk4 = Algorithm(nxn)
		
			yk0 = []; yk1 = []; yk2 = []; yk3 = []
		
			## omega k N is defined as omega 2*pi*k here, not omega -2*pi*k
			## as is seen often in FFT algorithms
			for k in range(0,N>>2):
				t0 = omega((k,N),zk4[k])
				t1 = omega((3*k,N),zpk4[k])
				s0 = t0 + t1
				s1 = img(N,t0 + negate(N,t1))
			
				yk0.append(uk2[k] + s0)
				yk2.append(uk2[k] + negate(N,s0))
				yk1.append(uk2[k+(N>>2)] + s1)
				yk3.append(uk2[k+(N>>2)] + negate(N,s1))
			
			return yk0+yk1+yk2+yk3
		
		self.ImportParseTree(Algorithm([Leaf(i) for i in range(0,self.N)]))

	def ParseBoolectorBFResult(self,fn):
		'''Parse a Boolector result and assign Wbase'''
		
		f = open(fn,'r')
		
		## Setup up zero Wbase
		Wbase = ModInt(0,self.N)
		for i in range(0,self.N):
			id = (self.N,i,0)
			self.weights[id] = Wbase
			id = (1,0,i)
			self.weights[id] = Wbase		
		
		
		for line in f.readlines():
			if line[:5] == "Wbase":
				res = line.split()
				nd = res[0]
				nd = nd.split("_")
				stride = int(nd[1])
				base = int(nd[2])
				Wstride = ModInt(int(nd[3]),self.N)
				lid = (stride,base,int(Wstride))
				rid = (stride,base,int(~Wstride))
				Wbase = ModInt(int(res[1],2),self.N)
				self.weights[lid] = Wbase
				self.weights[rid] = Wbase
		f.close()
		


