# FFTSearch: Genarating and Searching Families of FFT Algorithms
# Copyright (C) 2011 Steve Haynal, Heidi Haynal
#
# This file is part of FFTSearch.
#
# FFTSearch is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFTSearch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from FFTAlgorithm import *
from Node import *
from Constraints import *

import commands

## FFT flowgraph 
class Flowgraph:
	def __init__(self,m):
		
		self.m = m
		self.N = 2**m
		
		self.tpfg = None
		self.bpfg = None

		self.tl = None
		self.tr = None
		self.bl = None
		self.br = None
		
		self.inps = None
		self.outs = None
		
	def __str__(self):
		return "%s:%d" % (self.outs[0],self.m)
		
	def __repr__(self):
		return "Flowgraph(%d)" % self.m
		
	def __iter__(self):
		for sfg in self.BFS():
			for nd in sfg.Inps():
				yield nd

	def Inps(self):
		return iter(self.inps)
		
	def SubFlowGraphs(self):
		yield self.bl
		yield self.br
		
	def InpButterflies(self):
		nds = self.inps[:len(self.inps)/2]
		for nd in nds:
			yield nd.tpfg
	
	def OutButterflies(self):
		nds = self.outs[::2]
		for nd in nds:
			yield nd.bpfg
			
	def Butterflies(self):
		for inps in self.BFS():
			for bf in inps.InpButterflies():
				yield bf

	def BFS(self):
		yield self
		last = self
		for sfg in self.BFS():
			if self.m == 1:
				yield self.bl
				yield self.br
			else:
				for csfg in sfg.SubFlowGraphs():
					yield csfg
					last = csfg
				if last == sfg:
					return
				
	def DFS(self):
		yield self
		if self.m == 1:
			yield self.bl
			yield self.br
		else:
			for sfg in self.bl.DFS():
				yield sfg
			for sfg in self.br.DFS():
				yield sfg

	def SFGsOut2In(self,depth=1):
		if self.m > depth:
			for sfg in self.bl.SFGsOut2In(depth):
				yield sfg
			for sfg in self.br.SFGsOut2In(depth):
				yield sfg
		yield self
		
	def AllSFGs(self,seen=set([])):
		if self not in seen:
			if self.m > 1:
				for sfg in self.tl.AllSFGs(seen): yield sfg
				for sfg in self.tr.AllSFGs(seen): yield sfg
				for sfg in self.bl.AllSFGs(seen): yield sfg
				for sfg in self.br.AllSFGs(seen): yield sfg
			yield self
			seen.add(self)


	def InitTopFlowgraph(self):
		bfs = []
		def Do(xn,r):
			sz = len(xn)
			if sz == 1: return xn
			
			lr = zip(xn[:sz/2],xn[sz/2:])
			lxn = []; rxn = []
					
			ir = ~r
			for i,(lnd,rnd) in enumerate(lr):
				
				clnd = Node(sz>>1,i,r,ModInt(0,self.N))
				lnd.lc = clnd; rnd.lc = clnd
				clnd.lp = lnd; clnd.rp = rnd
				lxn.append(clnd)
				
				crnd = Node(sz>>1,i,ir,ModInt(0,self.N))
				lnd.rc = crnd; rnd.rc = crnd
				crnd.lp = lnd; crnd.rp = rnd
				rxn.append(crnd)
				
				bfs.append((clnd,crnd))

			return (Do(lxn,r>>1) + Do(rxn,(ir>>1)))

		self.inps = [Node(self.N,i,ModInt(0,self.N),ModInt(0,self.N)) for i in range(0,self.N)]
		self.outs = Do(self.inps,ModInt(0,self.N))

		## Build Butterfly subflowgraphs
		sfgs = []
		for (lnd,rnd) in bfs:
			##print "Building size",1
			sfg = Flowgraph(1)
			sfg.bl = lnd
			sfg.br = rnd
			sfg.tl = lnd.lp
			sfg.tr = lnd.rp
			lnd.bpfg = sfg
			rnd.bpfg = sfg
			lnd.lp.tpfg = sfg
			lnd.rp.tpfg = sfg
			sfg.inps = [lnd.lp,lnd.rp]
			sfg.outs = [lnd,rnd]
			sfgs.append(sfg)
		
		
		## Build larger subflowgraphs	
		while len(sfgs) != 0:
			nsfgs = []
			for sfg in sfgs:
				if sfg.bpfg == None:
					if sfg.tl.bpfg != None:
						##print "Building size",sfg.m+1
						if len(sfgs) == 4:
							## Use toplevel
							nsfg = self
						else:
							nsfg = Flowgraph(sfg.m+1)
						nsfg.tl = sfg.tl.bpfg
						nsfg.tr = sfg.tr.bpfg
						nsfg.bl = nsfg.tl.bl.tpfg
						nsfg.br = nsfg.tl.br.tpfg

						nsfg.tl.tpfg = nsfg
						nsfg.tr.tpfg = nsfg
						nsfg.bl.bpfg = nsfg
						nsfg.br.bpfg = nsfg
						if len(sfgs) != 4:
							nsfg.inps = sorted(nsfg.tl.inps + nsfg.tr.inps)
							nsfg.outs = sorted(nsfg.bl.outs + nsfg.br.outs)
						nsfgs.append(nsfg)
			sfgs = nsfgs
		
		## Add indexing
		lstride = self.N
		i = 0
		for nd in self:
			if lstride != nd.stride:
				i = 0
			lstride = nd.stride
			nd.i = i
			i += 1
					
			
	def GetID(self):
		return self.outs[0].id,m
		

	def UpdateTFsFromWeights(self):
		'''Synchronize TwiddleFactors to match Wbase'''
		for bf in self.Butterflies():
			bf.tl.ltfp = bf.bl.Wbase - bf.tl.Wbase
			bf.tl.rtfp = bf.br.Wbase - bf.tl.Wbase
			bf.tr.ltfp = bf.bl.rWbase() - bf.tr.Wbase
			bf.tr.rtfp = bf.br.rWbase() - bf.tr.Wbase
				
			if bf.bl.IsTerminalDescendent():
				bf.bl.ltfp = -bf.bl.Wbase
				bf.bl.rtfp = -bf.bl.Wbase
				bf.br.ltfp = -bf.br.Wbase
				bf.br.rtfp = -bf.br.Wbase
						
	def UpdateWeightsFromTFs(self):
		'''Synchornize Wbase to match TwiddleFactors'''
		for bf in self.Butterflies():
			bf.bl.Wbase = bf.tl.Wbase + bf.tl.ltfp
			bf.br.Wbase = bf.tl.Wbase + bf.tl.rtfp
			
			## Check that rWbase agree
			assert bf.bl.rWbase() == bf.tr.Wbase + bf.tr.ltfp, "Mismatching rWbase"
			assert bf.br.rWbase() == bf.tr.Wbase + bf.tr.rtfp, "Mismatching rWbase"
			
	def Cost(self,adds=True):
		'''Compute FLOP count of FFT flowgraph'''
		if adds:
			flops = self.m * self.N * 2
		else:
			flops = 0
		for nd in self:
			flops += nd.Cost()
		return flops
		
	def Validate(self):
		'''Validate the FFT flowgraph'''
		self.UpdateWeightsFromTFs()
		
		for nd in self:
			if nd.lp:
				assert nd.base == nd.lp.base, "Base not equal to left parent's base"
				assert nd.lp.base < nd.rp.base, "Bases not ordered with left parent as lowest"
				assert nd.stride == (nd.rp.base - nd.lp.base), "Stride calculation not correct"
				assert (nd.lp.base + (nd.lp.stride >> 1)) == nd.rp.base, "Not interleaving"
				assert nd.Wstride == (nd.rWbase() - nd.Wbase), "Incorrect Wstride"
				if not nd.lc:
					assert (nd.Wbase + nd.ltfp).v == 0, "Final Wbase is not zero"
		print "Validation completed with no errors"

			
	def SetFFTAlgorithm(self,ffta):
		if ffta.m != self.m:
			print "ERROR: Mismatching m in SetFFTAlgorithm",ffta.m,self.m
			return
			
		for nd in self:
			try:
				nd.Wbase = ffta.weights[nd.id]
			except:
				pass
			
		self.UpdateTFsFromWeights()

	def ExtractFFTAlgorithm(self):
		ffta = FFTAlgorithm(self.m)
		for nd in self:
			ffta.weights[nd.id] = nd.Wbase
		
		return ffta
		
	## Transformations
	def ScaleTF(self,s):
		for nd in self.inps:
			tf = nd.ltfp + s
			nd.SetTF(tf)
		for nd in self.outs:
			tf = nd.ltfp - s
			nd.SetTF(tf)
		
	## DIT algorithms have multiplication in top two rows
	## Scale these out
	def PrepareDIT(self):
		'''Prepare flowgraph for partitioning by ensuring no multiplication in top or bottom 2 rows'''
		if self.N <= 8: return
		for bf in self.br.InpButterflies():
			if bf.tl.ltfp.Norm() == bf.tr.ltfp.Norm():
				if bf.tl.ltfp.EqualRealImag():
					bf.ScaleTF(-bf.tl.ltfp)
		self.bl.PrepareDIT()


		 



