# FFTSearch: Genarating and Searching Families of FFT Algorithms
# Copyright (C) 2011 Steve Haynal, Heidi Haynal
#
# This file is part of FFTSearch.
#
# FFTSearch is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FFTSearch is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from scipy import *

## Container class for symbolic node data
class SymNode:
	def __init__(self,stride,base,Wbase,Wstride):
		self.stride = stride
		self.base = base
		self.Wbase = Wbase
		self.Wstride = Wstride

class TestFFT:
	def __init__(self,N):
		self.N = N
		self.m = int(log2(N))
		self.flops = 0
		
		## Create random complex input data
		self.data = (random.randn(self.N) +1j*random.randn(self.N)).astype(complex)
		
		## Create a golden result by calling SciPY's FFT
		self.gold = fft(self.data)
		
		## Create an array for symbolic data
		self.symdata = []
		for base in range(0,self.N):
			self.symdata.append(SymNode(self.N,base,0,0))
		
	def ComplexAdd(self,a,b):
		## 1 real addition for addition of real parts
		r = real(a) + real(b)
		self.flops += 1
		
		## 1 real addition for addition of imaginary parts
		i = imag(a) + imag(b)
		self.flops += 1
		
		## return complex number, no real addition or multiplication here
		return r+(i*1j)

	def ComplexSubtract(self,a,b):
		## Subtraction is just addition with free sign flipping
		## b is multiplied by -1 requiring no flops
		return self.ComplexAdd(a,-b)
			
	def ComplexMultiply(self,a,tf):
		r = real(a)
		i = imag(a)
		
		## Upper two bits of tf determine free multiplications by
		## 1,-1,i or -i
		freetf = tf >> (self.m-2)
		if freetf == 1:
			t = r
			r = i
			i = -t
		elif freetf == 2:
			r = -r
			i = -i
		elif freetf == 3:
			t = r
			r = -i
			i = t

		## Lower bits determine any multiplication requiring flops
		costtf = tf & ((self.N -1) >> 2)

		
		## costtf as complex number
		complexcosttf = exp(-2j*pi*(float32(costtf)/self.N))
		rtf = real(complexcosttf)
		itf = imag(complexcosttf)
		
		## Multiplication requiring only 4 flops
		if costtf == (self.N >> 3):
			rXrtf = r * rtf
			self.flops += 1
			
			iXrtf = i * rtf
			self.flops += 1
			
			r = rXrtf + iXrtf
			self.flops += 1
			
			i = iXrtf - rXrtf
			self.flops += 1

		## Multiplication requiring 6 flops			
		elif costtf != 0:
			
			rXrtf = r * rtf
			self.flops += 1
			
			iXrtf = i * rtf
			self.flops += 1
			
			rXitf = r * itf
			self.flops += 1
			
			iXitf = i * itf
			self.flops += 1			
						
			r = rXrtf - iXitf
			self.flops += 1
			
			i = iXrtf + rXitf
			self.flops += 1
						
			
		## return complex number, no real addition or multiplication here
		return r+(i*1j)							
		
	# li is index of left  nodes in size 2 butterfly
	# ri is index of right nodes in size 2 butterfly
	# lptfp is left parents twiddle factor power
	# rptfp is right parents twiddle factor power
	def BF(self,li,ri,lptfp,rptfp):
		
		l = self.ComplexMultiply(self.data[li],lptfp)
		r = self.ComplexMultiply(self.data[ri],rptfp)
		self.data[li] = self.ComplexAdd(l,r)
		self.data[ri] = self.ComplexSubtract(l,r)
		
		## Symbolic computation
		l = self.symdata[li]
		r = self.symdata[ri]
		
		## Graph structure checks
		## Strides must be the same
		if l.stride != r.stride:
			print "Graph Structure Error: Strides in butterfly are not the same",l.stride,r.stride
			
		## Left must have lower base
		if l.base > r.base:
			print "Graph Structure Error: Left base is not lower index",l.base,r.base
			
		## Must interleave
		if l.stride != ((r.base - l.base) * 2):
			print "Graph Structure Error: Right does not interleave with left",l.stride,r.base,l.base
			
		## Compute new left and right base and stride
		nbase = l.base
		nstride = r.base - l.base
			
			
		## Computation Checks
		## Weight strides must match
		if l.Wstride != r.Wstride:
			print "Symbolic Computation Error: Weight strides do not match",l.Wstride,r.Wstride
			
		## Compute new weights
		
		nlWbase = (l.Wbase + lptfp) % self.N
		nlrWbase = (r.Wbase + rptfp) % self.N
		nlWstride = (nlrWbase - nlWbase) % self.N
		self.symdata[li] = SymNode(nstride,nbase,nlWbase,nlWstride)
		## New left Wstride must be half of old Wstride
		if nlWstride != (l.Wstride / 2):
			print "Symbolic Computation Error: Incorrect new left Wstride",l.Wstride,nlWstride		
		
		nrWbase = (l.Wbase + lptfp) % self.N
		nrrWbase = (r.Wbase + rptfp + (self.N/2)) % self.N
		nrWstride = (nrrWbase - nrWbase) % self.N
		self.symdata[ri] = SymNode(nstride,nbase,nrWbase,nrWstride)		
		## New right Wstride must be half of old Wstride negated
		if nrWstride != ((l.Wstride /2) + (self.N/2)) % self.N:
			print "Symbolic Computation Error: Incorrect new right Wstride",l.Wstride,nrWstride
			

	# The formulation in the paper allows for weights to be applied
	# to a final result if necessary. This may incur multiplication cost
	def FW(self,i,tfp):
		
		l = self.ComplexMultiply(self.data[i],tfp)
		self.data[i] = l
		
	
	def BitReverseIndex(self,k):
		rk = 0
		for z in range(1,self.m):
			if k & 1: rk = rk | 1
			k =  k >> 1
			rk = rk << 1
		if k & 1: rk = rk | 1	
			
		return rk
		
		
		
	def Report(self):
		
		## Calculate Mean Absolute Error when compared to SciPy FFT
		## The accuracy of SciPy FFT may not be the best possible...
		mae = 0.0
		for k in range(0,self.N):
			gold = self.gold[k]
			## Output is generated in permuted order
			imp = self.data[self.BitReverseIndex(k)]
			mae = mae + abs(real(gold) - real(imp))
			mae = mae + abs(imag(gold) - imag(imp))
			
			## Do symbolic check
			symres = self.symdata[self.BitReverseIndex(k)]
			if symres.Wstride != k:
				print "ERROR: Final Wstride is incorrect for X(k)",k,symres.Wstride
			
		mae = mae / (2*self.N)
		
		print "N     =",self.N
		print "FLOPS =",self.flops			
		print "MAE   =",mae
		
		
		
		
